@@ -25,7 +25,6 @@ import inspect
 import textwrap
 import threading
 import contextlib
-import collections
 from io import BytesIO, StringIO
 import pandas as pd
 import requests
@@ -63,6 +62,7 @@ from .types import (
     DagInfo,
     DagInit,
     DagList,
+    DagPrettyNode,
     DagReload,
     DynamicResults,
     DynamicStatusResponse,
@@ -95,6 +95,7 @@ from .types import (
     NodeTiming,
     NodeTypes,
     NodeUserColumnsResponse,
+    PrettyResponse,
     PutNodeBlob,
     QueueMode,
     QueueStatsResponse,
@@ -1226,296 +1227,30 @@ class DagHandle:
             if success:
                 self.set_dynamic_error_message(None)
 
+    def _pretty(
+            self,
+            nodes_only: bool,
+            allow_unicode: bool) -> PrettyResponse:
+        return cast(PrettyResponse, self._client.request_json(
+            METHOD_GET, "/pretty", {
+                "dag": self.get_uri(),
+                "nodes_only": nodes_only,
+                "allow_unicode": allow_unicode,
+            }))
+
     def pretty(
-            self, nodes_only: bool = False, allow_unicode: bool = True) -> str:
-        nodes = [
-            self.get_node(node_id)
-            for node_id in sorted(self.get_nodes())
-        ]
-        already: Set[NodeHandle] = set()
-        order: List[NodeHandle] = []
-        outs: Dict[
-            NodeHandle,
-            List[Tuple[NodeHandle, str, str]],
-        ] = collections.defaultdict(list)
-        start_vert = "├" if allow_unicode else "|"
-        end_vert = "├" if allow_unicode else "|"
-        before_vert = "│" if allow_unicode else "|"
-        after_vert = "│" if allow_unicode else "|"
-        vert = "┤" if allow_unicode else "|"
-        corner_right = "┐" if allow_unicode else "\\"
-        corner_left = "┘" if allow_unicode else "/"
-        cont_right = "┬" if allow_unicode else "\\"
-        cont_left = "┴" if allow_unicode else "/"
-        cont_skip = "─" if allow_unicode else "-"
-        cont_vert = "│" if allow_unicode else "|"
-        cont = "┼" if allow_unicode else "-"
-        start_left = "└" if allow_unicode else "\\"
-        bar = "─" if allow_unicode else "-"
-        vsec = "│" if allow_unicode else "|"
-        vstart = "├" if allow_unicode else "|"
-        vend = "┤" if allow_unicode else "|"
-        hsec = "─" if allow_unicode else "-"
-        tl = "┌" if allow_unicode else "+"
-        tr = "┐" if allow_unicode else "+"
-        bl = "└" if allow_unicode else "+"
-        br = "┘" if allow_unicode else "+"
-        conn_top = "┴" if allow_unicode else "-"
-        conn_bottom = "┬" if allow_unicode else "-"
-        space = " "
-        prefix_len = 2 if nodes_only else 3
-        indent = space * prefix_len
+            self,
+            nodes_only: bool = False,
+            allow_unicode: bool = True) -> str:
+        return self._pretty(
+            nodes_only=nodes_only, allow_unicode=allow_unicode)["pretty"]
 
-        def topo(cur: NodeHandle) -> None:
-            if cur in already:
-                return
-            for in_key in sorted(cur.get_inputs()):
-                out_node, out_key = cur.get_input(in_key)
-                outs[out_node].append((cur, in_key, out_key))
-                topo(out_node)
-            already.add(cur)
-            order.append(cur)
-
-        for tnode in nodes:
-            topo(tnode)
-
-        in_states: Dict[NodeHandle, Dict[str, int]] = {}
-        order_lookup = {
-            node: pos for (pos, node) in enumerate(order)
-        }
-
-        def get_in_state(node: NodeHandle, key: str) -> int:
-            if node not in in_states:
-                in_states[node] = node.get_in_cursor_states()
-            return in_states[node].get(key, 0)
-
-        def draw_in_edges(
-                node: NodeHandle,
-                cur_edges: List[Tuple[Optional[NodeHandle], str, int]],
-                ) -> Tuple[List[Tuple[Optional[NodeHandle], str, int]], str]:
-            gap = 0
-            prev_gap = 0
-            new_edges: List[Tuple[Optional[NodeHandle], str, int]] = []
-            segs: List[str] = []
-            for edge in cur_edges:
-                in_node, in_key, cur_gap = edge
-                before_gap = cur_gap
-                if in_node == node:
-                    in_state = get_in_state(in_node, in_key)
-                    cur_str = f"{end_vert} {in_key} ({in_state}) "
-                    new_edges.append((None, in_key, cur_gap))
-                else:
-                    cur_str = before_vert if in_node is not None else ""
-                    cur_gap += gap
-                    gap = 0
-                    new_edges.append((in_node, in_key, cur_gap))
-                segs.append(f"{space * prev_gap}{cur_str}")
-                prev_gap = max(0, before_gap - len(cur_str))
-            while new_edges:
-                if new_edges[-1][0] is None:
-                    new_edges.pop()
-                else:
-                    break
-            return new_edges, "".join(segs)
-
-        def draw_out_edges(
-                node: NodeHandle,
-                cur_edges: List[Tuple[Optional[NodeHandle], str, int]],
-                ) -> Tuple[List[Tuple[Optional[NodeHandle], str, int]], str]:
-            new_edges: List[Tuple[Optional[NodeHandle], str, int]] = []
-            segs: List[str] = []
-            prev_gap = 0
-            for edge in cur_edges:
-                cur_node, _, cur_gap = edge
-                cur_str = after_vert if cur_node is not None else ""
-                segs.append(f"{space * prev_gap}{cur_str}")
-                new_edges.append(edge)
-                prev_gap = max(0, cur_gap - len(cur_str))
-            sout = sorted(
-                outs[node], key=lambda e: order_lookup[e[0]], reverse=True)
-            for (in_node, in_key, out_key) in sout:
-                cur_str = f"{start_vert} {out_key} "
-                in_state = get_in_state(in_node, in_key)
-                end_str = f"{end_vert} {in_key} ({in_state}) "
-                segs.append(f"{space * prev_gap}{cur_str}")
-                cur_gap = max(len(cur_str), len(end_str))
-                new_edges.append((in_node, in_key, cur_gap))
-                prev_gap = max(0, cur_gap - len(cur_str))
-            return new_edges, "".join(segs)
-
-        def draw() -> List[str]:
-            lines: List[str] = []
-            edges: List[Tuple[Optional[NodeHandle], str, int]] = []
-            for node in order:
-                top_gaps = [edge[2] for edge in edges[:-1]]
-                top_nodes = [edge[0] for edge in edges]
-                same_ids = [edge[0] == node for edge in edges]
-                empty_top = [edge[0] is None for edge in edges]
-                edges, in_line = draw_in_edges(node, edges)
-                in_line = in_line.rstrip()
-                if in_line:
-                    lines.append(f"{indent}{in_line}")
-                edges, out_line = draw_out_edges(node, edges)
-                bottom_gaps = [edge[2] for edge in edges[:-1]]
-                empty_bottom = [edge[0] is None for edge in edges]
-                new_bottom = [
-                    eix >= len(top_nodes) or (
-                        edge[0] is not None and top_nodes[eix] != edge[0]
-                    ) for (eix, edge) in enumerate(edges)
-                ]
-                line_indents: List[str] = []
-                started = False
-                had_same = False
-                highest_iix = -1
-                for (iix, top_gap) in enumerate(top_gaps):
-                    if same_ids[iix]:
-                        had_same = True
-                    if had_same and iix >= len(bottom_gaps):
-                        break
-                    if not line_indents:
-                        line_indents.append(indent)
-                    if empty_top[iix]:
-                        cur_connect = cont_skip if started else space
-                    elif iix >= len(empty_bottom) or empty_bottom[iix]:
-                        if started:
-                            cur_connect = cont_left
-                        else:
-                            cur_connect = start_left
-                        if len(bottom_gaps) < len(top_gaps):
-                            break
-                        started = True
-                    else:
-                        if started:
-                            cur_connect = cont_skip
-                        else:
-                            cur_connect = cont_vert
-                    cur_line = cont_skip if started else space
-                    gap_size = top_gap - len(cur_connect)
-                    line_indents.append(f"{cur_connect}{cur_line * gap_size}")
-                    highest_iix = iix
-                if line_indents:
-                    line_indents[-1] = line_indents[-1][:-len(indent)]
-                if nodes_only:
-                    mid = f" {node.get_short_status(allow_unicode)} "
-                else:
-                    mid = \
-                        f"{node.get_short_status(allow_unicode)} " \
-                        f"{node.get_name()}({node.get_type()}) " \
-                        f"{node.get_highest_chunk()}"
-                if len(mid) < prefix_len:
-                    mid = f"{mid}{space * (prefix_len - len(mid))}"
-                content = f"{vend if started else vsec}{mid}{vsec}"
-                node_line = f"{''.join(line_indents)}{content}"
-                top_indents: List[str] = []
-                bottom_indents: List[str] = []
-                for iix in range(highest_iix + 1):
-                    top_connect = space if empty_top[iix] else cont_vert
-                    has_bottom = iix >= len(empty_bottom) or empty_bottom[iix]
-                    bottom_connect = space if has_bottom else cont_vert
-                    top_gap_size = top_gaps[iix] - len(top_connect)
-                    bottom_gap_size = top_gaps[iix] - len(bottom_connect)
-                    if not top_indents:
-                        top_indents.append(indent)
-                    top_indents.append(f"{top_connect}{space * top_gap_size}")
-                    if not bottom_indents:
-                        bottom_indents.append(indent)
-                    bottom_indents.append(
-                        f"{bottom_connect}{space * bottom_gap_size}")
-                if top_indents:
-                    top_indents[-1] = top_indents[-1][:-len(indent)]
-                if bottom_indents:
-                    bottom_indents[-1] = bottom_indents[-1][:-len(indent)]
-                border_len = len(content) - len(tl) - len(tr)
-                top_border: List[str] = [tl] + [hsec] * border_len + [tr]
-                top_ix = len(indent)
-                for iix in range(highest_iix + 1, len(same_ids)):
-                    if top_ix >= len(top_border):
-                        break
-                    if same_ids[iix]:
-                        top_border[top_ix] = conn_top
-                    if iix >= len(top_gaps):
-                        break
-                    top_ix += top_gaps[iix]
-                bottom_border: List[str] = [bl] + [hsec] * border_len + [br]
-                bottom_ix = len(indent)
-                for iix in range(highest_iix + 1, len(new_bottom)):
-                    if bottom_ix >= len(bottom_border):
-                        break
-                    if new_bottom[iix]:
-                        bottom_border[bottom_ix] = conn_bottom
-                    if iix >= len(bottom_gaps):
-                        break
-                    bottom_ix += bottom_gaps[iix]
-                node_top = f"{''.join(top_indents)}{''.join(top_border)}"
-                node_bottom = \
-                    f"{''.join(bottom_indents)}{''.join(bottom_border)}"
-                total_gap_top = sum(top_gaps) - len(node_line)
-                total_gap_bottom = sum(bottom_gaps) - len(node_line)
-                if total_gap_bottom > total_gap_top:
-                    connector = corner_right
-                    more_gaps = bottom_gaps
-                    top_conn = space
-                    bottom_conn = cont_vert
-                else:
-                    connector = corner_left
-                    more_gaps = top_gaps
-                    top_conn = cont_vert
-                    bottom_conn = space
-                if total_gap_bottom == total_gap_top:
-                    connector = vert
-                    more_gaps = bottom_gaps
-                    top_conn = cont_vert
-                    bottom_conn = cont_vert
-                total_gap = max(total_gap_bottom, total_gap_top)
-                if total_gap >= -prefix_len:
-                    bar_len = total_gap + prefix_len
-                    full_bar = list(bar * bar_len)
-                    full_top = list(space * bar_len)
-                    full_bottom = list(space * bar_len)
-                    bar_ix = prefix_len - len(node_line)
-                    for (before_gap_ix, bar_gap) in enumerate(more_gaps):
-                        bar_ix += bar_gap
-                        if bar_ix < 0:
-                            continue
-                        if bar_ix >= len(full_bar):
-                            break
-                        gap_ix = before_gap_ix + 1
-                        if gap_ix < len(same_ids) and not same_ids[gap_ix]:
-                            mid_connector = cont_skip
-                            mid_top = cont_vert
-                            mid_bottom = cont_vert
-                        else:
-                            mid_connector = cont
-                            mid_top = cont_vert
-                            mid_bottom = cont_vert
-                        adj_ix = bar_ix - prefix_len
-                        if total_gap_bottom >= adj_ix > total_gap_top:
-                            mid_connector = cont_right
-                            mid_top = space
-                        elif total_gap_bottom < adj_ix <= total_gap_top:
-                            if not empty_top[gap_ix]:
-                                mid_connector = cont_left
-                            else:
-                                mid_top = space
-                            mid_bottom = space
-                        full_bar[bar_ix] = mid_connector
-                        full_top[bar_ix] = mid_top
-                        full_bottom[bar_ix] = mid_bottom
-                    node_line = \
-                        f"{node_line[:-len(vsec)]}{vstart}" \
-                        f"{''.join(full_bar)}{connector}"
-                    node_top = f"{node_top}{''.join(full_top)}{top_conn}"
-                    node_bottom = \
-                        f"{node_bottom}{''.join(full_bottom)}{bottom_conn}"
-                lines.append(node_top.rstrip())
-                lines.append(node_line)
-                lines.append(node_bottom.rstrip())
-                out_line = out_line.rstrip()
-                if out_line:
-                    lines.append(f"{indent}{out_line}")
-            return lines
-
-        return "\n".join(draw())
+    def pretty_obj(
+            self,
+            nodes_only: bool = False,
+            allow_unicode: bool = True) -> List[DagPrettyNode]:
+        return self._pretty(
+            nodes_only=nodes_only, allow_unicode=allow_unicode)["nodes"]
 
     def get_def(self, full: bool = True) -> DagDef:
         return cast(DagDef, self._client.request_json(
@@ -1582,12 +1317,16 @@ class DagHandle:
     def set_kafka_topic_partitions(
             self,
             num_partitions: int,
-            large_input_retention: bool = False) -> KafkaTopics:
+            postfix: str = "",
+            large_input_retention: bool = False,
+            no_output: bool = False) -> KafkaTopics:
         return cast(KafkaTopics, self._client.request_json(
             METHOD_POST, "/kafka_topics", {
                 "dag": self.get_uri(),
                 "num_partitions": num_partitions,
+                "postfix": postfix,
                 "large_input_retention": large_input_retention,
+                "no_output": no_output,
             }))
 
     def post_kafka_objs(self, input_objs: List[Any]) -> List[str]:
@@ -1601,11 +1340,15 @@ class DagHandle:
         ]
         return self.post_kafka_msgs(bios)
 
-    def post_kafka_msgs(self, input_data: List[BytesIO]) -> List[str]:
+    def post_kafka_msgs(
+            self,
+            input_data: List[BytesIO],
+            postfix: str = "") -> List[str]:
         names = [f"file{pos}" for pos in range(len(input_data))]
         res = cast(KafkaMessage, self._client.request_json(
             METHOD_FILE, "/kafka_msg", {
                 "dag": self.get_uri(),
+                "postfix": postfix,
             }, files=dict(zip(names, input_data))))
         msgs = res["messages"]
         return [msgs[key] for key in names]
@@ -1646,20 +1389,25 @@ class DagHandle:
             return None
         return merge_ctype(res, ctype)
 
-    def get_kafka_offsets(self, alive: bool) -> KafkaOffsets:
+    def get_kafka_offsets(
+            self, alive: bool, postfix: Optional[str] = None) -> KafkaOffsets:
+        args = {
+            "dag": self.get_uri(),
+            "alive": int(alive),
+        }
+        if postfix is not None:
+            args["postfix"] = postfix
         return cast(KafkaOffsets, self._client.request_json(
-            METHOD_GET, "/kafka_offsets", {
-                "dag": self.get_uri(),
-                "alive": int(alive),
-            }))
+            METHOD_GET, "/kafka_offsets", args))
 
     def get_kafka_throughput(
             self,
+            postfix: Optional[str] = None,
             segment_interval: float = 120.0,
             segments: int = 5) -> KafkaThroughput:
         assert segments > 0
         assert segment_interval > 0.0
-        offsets = self.get_kafka_offsets(alive=False)
+        offsets = self.get_kafka_offsets(postfix=postfix, alive=False)
         now = time.monotonic()
         measurements: List[Tuple[int, int, int, float]] = [(
             offsets["input"],
@@ -1672,7 +1420,7 @@ class DagHandle:
             while now - prev < segment_interval:
                 time.sleep(max(0.0, segment_interval - (now - prev)))
                 now = time.monotonic()
-            offsets = self.get_kafka_offsets(alive=False)
+            offsets = self.get_kafka_offsets(postfix=postfix, alive=False)
             measurements.append((
                 offsets["input"],
                 offsets["output"],
